import ipaddress
from typing import List

from django.core.validators import MinValueValidator, MaxValueValidator, RegexValidator
from django.db import models
from django.db.models.signals import pre_save, pre_delete
from django.dispatch import receiver
from django.utils.translation import ugettext_lazy as _

from django_wireguard import settings
from django_wireguard.signals import interface_created, interface_deleted
from django_wireguard.utils import clean_comma_separated_list
from django_wireguard.validators import validate_private_ipv4, validate_wireguard_private_key, \
    validate_wireguard_public_key, validate_allowed_ips
from django_wireguard.wireguard import WireGuard, PrivateKey

__all__ = ('WireguardInterface', 'WireguardPeer')


class WireguardInterface(models.Model):
    name = models.CharField(max_length=100,
                            validators=[RegexValidator(r'^[a-zA-Z0-9]+$',
                                                       _("Interface Name must be a string of alphanumeric chars."))],
                            unique=True,
                            verbose_name=_("Interface Name"))
    listen_port = models.PositiveIntegerField(validators=[MinValueValidator(1), MaxValueValidator(65535)],
                                              verbose_name=_("Listen Port"))
    address = models.TextField(blank=True,
                               validators=[validate_allowed_ips],
                               verbose_name=_("Interface addresses."))
    private_key = models.CharField(max_length=64,
                                   blank=True,
                                   validators=[validate_wireguard_private_key],
                                   verbose_name=_("Private Key (leave empty to auto generate)"))

    class Meta:
        verbose_name = _("WireGuard Interface")
        verbose_name_plural = _("WireGuard Interfaces")

    @property
    def public_key(self) -> str:
        return str(PrivateKey(self.private_key).public_key())

    @property
    def wg(self) -> WireGuard:
        interface, created = WireGuard.get_or_create_interface(self.name)
        if created:
            interface_created.send(sender=self.__class__, instance=self)
        return interface

    def __repr__(self):
        return f"{self._meta.verbose_name} {self.name}"

    def __str__(self):
        return f"{self.name} - {self.address}"

    def get_address_list(self):
        return clean_comma_separated_list(self.address)

    def get_endpoint(self):
        return f"{settings.WIREGUARD_ENDPOINT}:{self.listen_port}"


class WireguardPeer(models.Model):
    interface = models.ForeignKey(WireguardInterface,
                                  on_delete=models.CASCADE,
                                  related_name='peers',
                                  verbose_name=_("Interface"))

    name = models.CharField(max_length=100,
                            blank=False)
    private_key = models.CharField(max_length=64,
                                   null=True,
                                   blank=True,
                                   validators=[validate_wireguard_private_key],
                                   verbose_name=_("Peer's Private Key"))
    public_key = models.CharField(max_length=64,
                                  unique=True,
                                  blank=True,
                                  validators=[validate_wireguard_public_key],
                                  verbose_name=_("Peer's Public Key"))
    dns = models.TextField(blank=True, verbose_name=_("DNS"),
                           validators=[validate_allowed_ips],
                           help_text=_("Comma separated list."))
    # Peer's IP inside the VPN network
    address = models.CharField(validators=[validate_private_ipv4],
                               max_length=20,
                               blank=True,
                               verbose_name=_("Address"))
    interface_allowed_ips = models.TextField(validators=[validate_allowed_ips],
                                             blank=True,
                                             verbose_name=_("Interface Allowed IPs"),
                                             help_text=_("One per line"))
    allowed_ips = models.TextField(validators=[validate_allowed_ips],
                                   blank=True,
                                   verbose_name=_("Allowed IPs"),
                                   help_text=_("Comma separated list."))
    interface_persistent_keepalive = models.PositiveIntegerField(blank=True,
                                                                 default=0,
                                                                 verbose_name=_("Interface Persistent Keepalive"))
    persistent_keepalive = models.PositiveIntegerField(blank=True,
                                                       default=0,
                                                       verbose_name=_("Persistent Keepalive"))

    class Meta:
        verbose_name = _("WireGuard Peer")
        verbose_name_plural = _("WireGuard Peers")
        unique_together = ('interface', 'name')

    def __repr__(self):
        return f"{self._meta.verbose_name} {self.name} - interface {self.interface}"

    def __str__(self):
        return f"{self.name}@{self.interface.name} - {self.address}"

    def get_dns_list(self) -> List[str]:
        return clean_comma_separated_list(self.dns)

    def get_allowed_ips(self) -> List[str]:
        return clean_comma_separated_list(self.allowed_ips)

    def get_interface_allowed_ips(self) -> List[str]:
        values = clean_comma_separated_list(self.interface_allowed_ips)
        if self.address:
            values.append(self.address)
        return values

    def get_config(self) -> str:
        """
        Generate WireGuard configuration for peer as string.

        If ``WIREGUARD_STORE_PRIVATE_KEYS`` is False, the string
        ``<INSERT-PRIVATE-KEY-FOR:{public_key}>`` will be placed as PrivateKey.

        :return: Peer configuration as string.
        """
        private_key = self.private_key or _('<INSERT-PRIVATE-KEY-FOR:%(pubkey)s>') % {'pubkey': self.public_key}

        config = f"[Interface]\n" \
                 f"Address={self.address}/32\n" \
                 f"PrivateKey={private_key}\n"

        if self.dns:
            config += f"DNS={self.dns}\n"

        config += f"[Peer]\n" \
                  f"Endpoint={self.interface.get_endpoint()}\n" \
                  f"PublicKey={self.interface.public_key}\n" \
                  f"AllowedIPs={self.allowed_ips}"

        return config


@receiver(pre_save, sender=WireguardInterface)
def sync_wireguard_interface(sender, **kwargs):
    interface = kwargs['instance']
    if not interface.private_key:
        interface.private_key = str(PrivateKey.generate())

    interface.wg.set_interface(private_key=interface.private_key,
                               listen_port=interface.listen_port)

    interface.wg.set_ip_addresses(*interface.get_address_list())


@receiver(pre_save, sender=WireguardPeer)
def sync_wireguard_peer(sender, **kwargs):
    peer: WireguardPeer = kwargs['instance']
    interface = peer.interface

    if not peer.address:
        for address in interface.wg.get_ip_addresses():
            address = ipaddress.IPv4Interface(address)
            interface_address = str(address.ip)
            subnet = address.network

            # fetch other peers for checks
            peers = WireguardPeer.objects.all().only('address')

            # auto assign IP address
            peer_addresses = [peer.address for peer in peers]
            peer_addresses.append(interface_address)
            if not peer.address:
                for host in subnet.hosts():
                    if str(host) not in peer_addresses:
                        peer.address = str(host)
                        break
                else:
                    continue
            break
        else:
            raise RuntimeWarning("WireGuard interface's subnets have no available IP left")

    if not peer.private_key and not peer.public_key:
        if settings.WIREGUARD_STORE_PRIVATE_KEYS:
            peer.private_key = str(PrivateKey.generate())
        else:
            # this shouldn't happen with standard form, but manual creation could lead here
            raise RuntimeError("WIREGUARD_STORE_PRIVATE_KEYS is False, yet no key has been passed.")

    # generate and store public key if the private key is provided
    if peer.private_key:
        peer.public_key = str(PrivateKey(peer.private_key).public_key())

    # update/create the wireguard peer
    interface.wg.set_peer(peer.public_key,
                          *peer.get_interface_allowed_ips())


@receiver(pre_delete, sender=WireguardInterface)
def delete_interface(sender, **kwargs):
    interface: WireguardInterface = kwargs['instance']
    interface.wg.delete()
    interface_deleted.send(sender, instance=interface)


@receiver(pre_delete, sender=WireguardPeer)
def delete_peer(sender, **kwargs):
    peer: WireguardPeer = kwargs['instance']
    peer.interface.wg.remove_peers(peer.public_key)



from django.core.management.base import BaseCommand

from django_wireguard.utils import purge_private_keys


class Command(BaseCommand):
    help = 'Purge Private Key information from all stored peers.'

    def handle(self, *args, **options):
        count = purge_private_keys()
        if count > 0:
            self.stderr.write(self.style.SUCCESS(f"Removed all {count} Private Keys from database."))
        else:
            self.stderr.write(self.style.NOTICE(f"No Private Key found in database."))



from typing import List


def purge_private_keys() -> int:
    """
    Purge all Private Key from database.

    :return: Removed keys count.
    """
    from django_wireguard.models import WireguardPeer

    peers = WireguardPeer.objects.filter(private_key__isnull=False)
    if peers.exists():
        peers.update(private_key=None)

    return peers.count()


def clean_comma_separated_list(value: str) -> List[str]:
    """
    Clean comma separated list of values.

    Removes all spaces and newlines (CR, CRLF, LF chars).
    Removes trailing commas.
    Returns empty list if string is empty after evaluation.

    :value: comma-separated list of values
    :type value: str
    :return: list of values separated and trimmed from input
    :rtype: list
    """
    values = (value
              .replace(' ', '')
              .replace('\r', '')
              .replace('\n', '')
              .rstrip(',')
              .split(','))

    if values == ['']:
        return []

    return values



import ipaddress
from io import StringIO
from unittest import mock

from django.core.management import call_command
from django.test import TestCase, tag

from django_wireguard.models import WireguardInterface, WireguardPeer
from django_wireguard.signals import interface_created, interface_deleted
from django_wireguard.tests.helpers import SignalHelper
from django_wireguard.utils import purge_private_keys, clean_comma_separated_list
from django_wireguard.wireguard import WireGuard, WireGuardException, PrivateKey, PublicKey


class TestWireguardModule(TestCase):
    interface_name = 'testInterface'

    @tag('net')
    def test_interface_not_exists(self):
        with self.assertRaises(WireGuardException):
            WireGuard(self.interface_name)

    def test_private_key(self):
        private_key = "cIfHrTxl4H+ojFn1LIzMHM+IpxOmbKWjhCtW/vd/YFQ="
        public_key = "aEMb9hff7fiVV2f2avti+b37S31ZaBLxfMLgiLD7z14="

        with self.assertRaises(TypeError):
            # noinspection PyArgumentList
            PrivateKey()

        obj = PrivateKey(private_key)
        self.assertIsInstance(obj, PrivateKey)
        self.assertEqual(str(obj), private_key)
        self.assertIsInstance(obj.public_key(), PublicKey)
        self.assertEqual(str(obj.public_key()), public_key)

        obj = PrivateKey.generate()
        self.assertIsInstance(obj, PrivateKey)

    def test_public_key(self):
        public_key = "aEMb9hff7fiVV2f2avti+b37S31ZaBLxfMLgiLD7z14="

        with self.assertRaises(TypeError):
            # noinspection PyArgumentList
            PublicKey()

        obj = PublicKey(public_key)
        self.assertIsInstance(obj, PublicKey)
        self.assertEqual(str(obj), public_key)


class TestWireguardInterface(TestCase):
    interface_name = 'testInterface'
    test_port = 4321
    test_addresses = '10.100.20.1/24,10.100.0.1/32'

    @tag('net')
    def test_interface_creation(self):
        interface = WireguardInterface.objects.create(name=self.interface_name,
                                                      listen_port=self.test_port,
                                                      address=self.test_addresses)

        try:
            wg = WireGuard(self.interface_name)
        except WireGuardException:
            self.fail("IPROUTE WireGuard Interface not created.")

        ip_addresses = wg.get_ip_addresses()
        for address in interface.get_address_list():
            ip = str(ipaddress.IPv4Interface(address))
            self.assertIn(ip, ip_addresses)

        # delete interface manually to trigger interfaces cleanup
        interface.delete()


class TestSignals(TestCase):
    interface_name = 'testInterface'
    test_port = 4321
    test_addresses = '10.100.20.1/24,10.100.0.1/32'

    @tag('signals')
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__ipr',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__wg',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__connect_backend',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    @mock.patch('django_wireguard.wireguard.WireGuard.get_or_create_interface',
                mock.MagicMock(return_value=(mock.MagicMock(return_value=True), True)))
    def test_interface_created_signal_sent(self):
        """
        Test ``interface_created`` signal is sent upon WireguardInterface instance creation.
        """
        with SignalHelper(interface_created, WireguardInterface) as signal:
            interface = WireguardInterface.objects.create(name=self.interface_name,
                                                          listen_port=self.test_port,
                                                          address=self.test_addresses)
            self.assertTrue(signal.signal_sent)
            self.assertIs(signal.signal_sender, WireguardInterface)
            self.assertEqual(signal.signal_kwargs['instance'], interface)

    @tag('signals')
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__ipr',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__wg',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__connect_backend',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    def test_interface_deleted_signal_sent(self):
        """
        Test ``interface_deleted`` signal is sent upon WireguardInterface instance deleted.

        Mocks ``WireGuard.create_interface`` and ``WireGuard.delete`` methods to avoid the need for
        Wireguard capabilities enabled.
        """
        with SignalHelper(interface_deleted, WireguardInterface) as signal:
            interface = WireguardInterface.objects.create(name=self.interface_name,
                                                          listen_port=self.test_port,
                                                          address=self.test_addresses)
            interface.delete()
            self.assertTrue(signal.signal_sent)
            self.assertIs(signal.signal_sender, WireguardInterface)
            self.assertEqual(signal.signal_kwargs['instance'], interface)


class TestWireguardPeer(TestCase):
    """
    Test WireguardPeer models creation and cleanup utils.

    Mocks ``Wireguard.create_interface`` methods to avoid the need for
    Wireguard capabilities enabled.
    """
    interface_name = 'testInterface'
    test_port = 4321
    test_addresses = '10.100.20.1/24,10.100.0.1/32'

    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__ipr',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__wg',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__connect_backend',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    def setUp(self):
        self.interface = WireguardInterface.objects.create(name=self.interface_name,
                                                           listen_port=self.test_port,
                                                           address=self.test_addresses)

    @tag('utils')
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__ipr',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__wg',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__connect_backend',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    def test_key_cleanup(self):
        WireguardPeer.objects.create(interface=self.interface,
                                     address='10.100.20.2',
                                     private_key='cIfHrTxl4H+ojFn1LIzMHM+IpxOmbKWjhCtW/vd/YFQ=')
        self.assertEqual(1, WireguardPeer.objects.filter(private_key__isnull=False).count())
        purge_private_keys()
        self.assertEqual(0, WireguardPeer.objects.filter(private_key__isnull=False).count())


class TestManagementCommands(TestCase):
    interface_name = 'testInterface'
    test_port = 4321
    test_addresses = ['10.100.20.1/24', '10.100.0.1/32']

    @tag('command')
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__ipr',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__wg',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__connect_backend',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    def test_setup_command_creation(self):
        out = StringIO()
        err = StringIO()
        call_command('setup_interface',
                     self.interface_name,
                     "--listen-port", self.test_port,
                     "--address", *self.test_addresses,
                     stdout=out, stderr=err)
        self.assertIn("create", err.getvalue())
        self.assertIn(self.interface_name, err.getvalue())

        interface = WireguardInterface.objects.filter(name=self.interface_name,
                                                      listen_port=self.test_port)
        self.assertTrue(interface.exists())

        addresses = interface.first().get_address_list()
        for address in self.test_addresses:
            self.assertIn(address, addresses)

    @tag('command')
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__ipr',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__wg',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    @mock.patch('django_wireguard.wireguard.WireGuard._WireGuard__connect_backend',
                mock.MagicMock(return_value=mock.MagicMock(return_value=False)))
    def test_setup_command_update(self):
        WireguardInterface.objects.create(
            name=self.interface_name,
            listen_port=self.test_port,
            address=self.test_addresses[0]
        )

        out = StringIO()
        err = StringIO()
        call_command('setup_interface',
                     self.interface_name,
                     "--listen-port", self.test_port,
                     "--address", *self.test_addresses,
                     stdout=out, stderr=err)
        self.assertIn("update", err.getvalue())
        self.assertIn(self.interface_name, err.getvalue())

        interface = WireguardInterface.objects.filter(name=self.interface_name,
                                                      listen_port=self.test_port)
        self.assertTrue(interface.exists())

        addresses = interface.first().get_address_list()
        for address in self.test_addresses:
            self.assertIn(address, addresses)


class TestUtils(TestCase):
    @tag('utils')
    def test_comma_separated_list_clean(self):
        """
        Test ``utils.clean_comma_separated_list`` edge cases.
        """
        expected = ['1', '2', '3']
        res = clean_comma_separated_list("1,2,3")
        self.assertEqual(expected, res, "Clean function not correctly evaluating values at comma")
        res = clean_comma_separated_list("1\n,\n2,3")
        self.assertEqual(expected, res, "Clean function not correctly evaluating LF")
        res = clean_comma_separated_list("1\r\n,\r2,3")
        self.assertEqual(expected, res, "Clean function not correctly evaluating CR")
        res = clean_comma_separated_list("1 ,  2 ,  3 ")
        self.assertEqual(expected, res, "Clean function not correctly evaluating spaces")
        res = clean_comma_separated_list("1,2,3,,")
        self.assertEqual(expected, res, "Clean function not correctly evaluating trailing commas")
        res = clean_comma_separated_list("1,\n  2\r, 3\r\n,\n,")
        self.assertEqual(expected, res, "Clean function not correctly evaluating mixed cases")

        res = clean_comma_separated_list(",\n \r,,, \r\n,\n,")
        self.assertEqual([], res, "Clean function not correctly evaluating empty values")



from django.contrib import admin
from django.utils.safestring import mark_safe
from wagtail.contrib.modeladmin.options import ModelAdmin, modeladmin_register

from django_wireguard import settings
from django_wireguard.models import WireguardPeer, WireguardInterface
from django_wireguard.forms import WireguardPeerForm


class WireguardPeerInlineAdmin(admin.TabularInline):
    model = WireguardPeer
    show_change_link = True


@modeladmin_register
class WireguardInterfaceAdmin(ModelAdmin):
    model = WireguardInterface
    menu_label = 'Wireguard Interfaces'
    menu_icon = 'lock'
    list_display = ('name', 'address', 'listen_port', 'public_key',)
    search_fields = ('name', 'address', 'listen_port')
    add_to_settings_menu = settings.WIREGUARD_WAGTAIL_SHOW_IN_SETTINGS

    inspect_view_fields = ('name', 'address')


@modeladmin_register
class WireguardPeerAdmin(ModelAdmin):
    model = WireguardPeer
    form = WireguardPeerForm
    menu_label = 'Wireguard Peers'
    menu_icon = 'lock'
    list_display = ('name', 'address', 'public_key',)
    list_filter = ()
    search_fields = ('name', 'address')
    add_to_settings_menu = settings.WIREGUARD_WAGTAIL_SHOW_IN_SETTINGS

    inspect_view_enabled = True
    inspect_template_name = 'django_wireguard/wireguardpeer_inspect.html'
    inspect_view_extra_js = ['js/qrcode.min.js', 'js/inject_qrcode.js']
    inspect_view_fields = ('name', 'address')

    def config(self, obj):
        return mark_safe(f'<pre>{obj.get_config()}</pre>')
    config.short_description = 'Config'



"""test URLconf
"""
from django.contrib import admin
from django.urls import path, include
from wagtail.admin import urls as wagtailadmin_urls

urlpatterns = [
    path('admin/', admin.site.urls),
    path('wagtail/', include(wagtailadmin_urls)),
]



from typing import Union, Optional

from django.db.models import QuerySet

from django_wireguard.models import WireguardInterface


def sync_wireguard_interfaces(queryset: Optional[Union[QuerySet, WireguardInterface]] = None):
    """
    Sync database WireguardInterface queryset or instance to WireGuard devices on the system.

    :param queryset: WireguardInterfaces to sync with the system
    :type queryset: WireguardInterface queryset or instance
    """
    if queryset is None:
        queryset = WireguardInterface.objects.all()
    elif isinstance(queryset, WireguardInterface):
        queryset = [queryset]

    for interface in queryset:
        interface.wg.set_interface(private_key=interface.private_key,
                                   listen_port=interface.listen_port)

        if interface.address:
            interface.wg.set_ip_addresses(*interface.get_address_list())

        for peer in interface.peers.all():
            # update/create the wireguard peer
            interface.wg.set_peer(peer.public_key,
                                  *peer.get_interface_allowed_ips())



import re

from django.core.management.base import BaseCommand
from django.core.validators import RegexValidator

from django_wireguard.models import WireguardInterface


class Command(BaseCommand):
    help = 'Setup WireGuard interface'

    def add_arguments(self, parser):
        parser.add_argument('interface_name',
                            type=str,
                            help="Interface Name")
        parser.add_argument('--listen-port',
                            nargs='?',
                            type=int,
                            default=1194,
                            help="Listen Port for the interface")
        parser.add_argument('--private-key',
                            nargs='?',
                            type=str,
                            help="Base64 encoded Private Key")
        parser.add_argument('--address',
                            nargs='*',
                            type=str,
                            help="Interface address(es)")

    def handle(self, *args, **options):
        if not re.match(r'^[a-zA-Z0-9]+$', options['interface_name']):
            self.stderr.write(self.style.ERROR("Interface Name must be a string of alphanumeric characters."))
            return

        interface = WireguardInterface.objects.filter(name=options['interface_name'])

        address = ','.join(options['address'] or [])

        if interface.exists():
            if options['private_key']:
                interface.update(listen_port=options['listen_port'],
                                 private_key=options['private_key'],
                                 address=address)
            else:
                interface.update(listen_port=options['listen_port'],
                                 address=address)
            self.stderr.write(self.style.SUCCESS(f"Interface updated: {interface.first().name}.\n"))
        else:
            interface = WireguardInterface.objects.create(name=options['interface_name'],
                                                          listen_port=options['listen_port'],
                                                          private_key=options['private_key'],
                                                          address=address)

            self.stderr.write(self.style.SUCCESS(f"Interface created: {interface.name}.\n"))



from django.core.management.base import BaseCommand

from django_wireguard.models import WireguardPeer


class Command(BaseCommand):
    help = 'Create a WireGuard Peer'

    def add_arguments(self, parser):
        group = parser.add_mutually_exclusive_group()
        group.add_argument('pubkey',
                           type=str,
                           nargs='*',
                           help="Peer Public Key(s)")
        group.add_argument('--all',
                           action='store_true',
                           help="Delete all peers")

    def handle(self, *args, **options):
        public_keys = options['pubkey']
        delete_all = options['all']

        if delete_all:
            peers = WireguardPeer.objects.all()
        else:
            peers = WireguardPeer.objects.filter(public_key__in=public_keys)

        if not peers.exists():
            self.stderr.write(self.style.NOTICE("No peer deleted."))
            return

        peer_names = '\n'.join(peer.name for peer in peers)
        peers.delete()
        self.stderr.write(self.style.SUCCESS(f"Deleted peers:\n"
                                             f"--------------\n"
                                             f"{peer_names}\n"))



"""
Custom Signals sent on WireGuard interface status changes.
"""

from django.dispatch import Signal

interface_created = Signal(providing_args=('instance',))
"""Signal sent to emulate WireGuard's PostUp configuration option

Provides instance argument (sender instance).
"""

interface_deleted = Signal(providing_args=('instance',))
"""Signal sent to emulate WireGuard's PostDown configuration option

Provides instance argument (sender instance).
"""



from typing import List

from django.dispatch import receiver


class SignalHelper:
    def __init__(self, signal_class, sender=None):
        self.signal_class = signal_class
        self.expected_sender = sender
        self.signal_sent = False
        self.signal_sender = None
        self.signal_args = ()
        self.signal_kwargs = {}

    def _handler(self, sender, *args, **kwargs):
        self.signal_sent = True
        self.signal_sender = sender
        self.signal_args = args
        self.signal_kwargs = kwargs

    def __enter__(self):
        self.signal_class.connect(self._handler, sender=self.expected_sender)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.signal_class.disconnect(self._handler, sender=self.expected_sender)



"""
Django base settings for SimpleVPN project.
"""

import os


BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))

SECRET_KEY = '&e1tf5lh6(le8s2l2r_rlmp24n@hki=+^8+6+^=_521&n2n66q'

DEBUG = True

ALLOWED_HOSTS = ['*']

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

INSTALLED_APPS = [
    'wagtail.contrib.forms',
    'wagtail.contrib.redirects',
    'wagtail.embeds',
    'wagtail.sites',
    'wagtail.users',
    'wagtail.snippets',
    'wagtail.documents',
    'wagtail.images',
    'wagtail.search',
    'wagtail.admin',
    'wagtail.contrib.modeladmin',
    'wagtail.core',

    'modelcluster',
    'taggit',

    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    'django_wireguard',
    'django_wireguard.wagtail',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'django_wireguard.tests.testapp.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages'
            ],
        },
    },
]

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

STATIC_URL = '/static/'

WAGTAIL_SITE_NAME = "django_wireguard_test"



from django import forms
from django.utils.translation import ugettext_lazy as _

from django_wireguard import settings
from django_wireguard.models import WireguardPeer


class WireguardPeerForm(forms.ModelForm):
    class Meta:
        model = WireguardPeer
        help_texts = {}
        if settings.WIREGUARD_STORE_PRIVATE_KEYS:
            fields = '__all__'
            help_texts.update({
              'private_key': _("Leave blank to generate one automatically. "
                               "You may specify only the Public Key for good security, but you'll have to generate "
                               "your own Private Key to place inside the configuration."),
              'public_key': _("Leave empty if Private Key is used instead (or left empty too for auto generation).")
            })
        else:
            exclude = ('private_key',)
            help_texts.update({
              'public_key': _("Generate a Private Key first, then place here the associated Public Key. "
                              "To learn more, follow the official WireGuard guide: "
                              "https://www.wireguard.com/quickstart/#key-generation")
            })

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['public_key'].required = not settings.WIREGUARD_STORE_PRIVATE_KEYS

    def clean(self):
        cleaned_data = super().clean()
        private_key = cleaned_data.get('private_key')
        public_key = cleaned_data.get('public_key')
        if private_key and public_key:
            self.add_error('public_key', _("Public Key is not required when Private Key is specified."))
        
        if not settings.WIREGUARD_STORE_PRIVATE_KEYS and not public_key:
            self.add_error('public_key', _("Enable WIREGUARD_STORE_PRIVATE_KEYS for "
                                           "automatic key handling, but keep in mind that doing so impacts security "
                                           "considerably."))



import ipaddress
from typing import Union

from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import ValidationError
from django_wireguard.wireguard import PrivateKey, PublicKey


def validate_private_ipv4(value: str):
    """
    Validate private IPv4 address.

    :param value: Value to validate
    :type value: str
    :raises: ValidationError if value provided is not a valid IPv4 address
    """
    try:
        ip_address = ipaddress.IPv4Address(value)
        if not ip_address.is_private or ip_address.is_unspecified:
            raise ValueError
    except (ValueError, ipaddress.AddressValueError):
        raise ValidationError(
            _('%(value)s is not a valid private IP Address.'),
            params={'value': value},
        )


def validate_allowed_ips(value: str):
    """
    Validate comma-separated list of allowed_ips (IPv4 interfaces).

    :param value: Value to validate
    :type value: str
    :raises: ValidationError if value provided is not a valid list of IPv4 interfaces
    """
    for ip in value.split(','):
        try:
            ip_address = ipaddress.IPv4Interface(ip)
            if not ip_address.is_private or ip_address.is_unspecified:
                raise ValueError
        except (ValueError, ipaddress.AddressValueError):
            raise ValidationError(
                _('%(value)s is not a valid private IP Address.'),
                params={'value': value},
            )


def validate_wireguard_private_key(value: str):
    """
    Validate base64 encoded private key.

    :param value: Value to validate
    :raises: ValidationError if value provided is not a valid private key
    """
    try:
        PrivateKey(value)
    except ValueError:
        raise ValidationError(
            _('The value specified is not a valid WireGuard Private Key.'),
        )


def validate_wireguard_public_key(value):
    """
    Validate base64 encoded public key.

    :param value: Value to validate
    :raises: ValidationError if value provided is not a valid public key
    """
    try:
        PublicKey(value)
    except ValueError:
        raise ValidationError(
            _('The value specified is not a valid WireGuard Public Key.'),
        )



# Generated by Django 2.2.7 on 2020-11-03 00:04

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django_wireguard.validators


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='WireguardInterface',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, unique=True, validators=[django.core.validators.RegexValidator('^[A-z0-9]+$', 'Interface Name must be a string of alphanumeric chars.')], verbose_name='Interface Name')),
                ('listen_port', models.PositiveIntegerField(validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(65535)], verbose_name='Listen Port')),
                ('address', models.TextField(blank=True, validators=[django_wireguard.validators.validate_allowed_ips], verbose_name='Interface addresses.')),
                ('private_key', models.CharField(blank=True, max_length=64, validators=[django_wireguard.validators.validate_wireguard_private_key], verbose_name='Private Key (leave empty to auto generate)')),
            ],
            options={
                'verbose_name': 'WireGuard Interface',
                'verbose_name_plural': 'WireGuard Interfaces',
            },
        ),
        migrations.CreateModel(
            name='WireguardPeer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('private_key', models.CharField(blank=True, max_length=64, null=True, validators=[django_wireguard.validators.validate_wireguard_private_key], verbose_name="Peer's Private Key")),
                ('public_key', models.CharField(blank=True, max_length=64, unique=True, validators=[django_wireguard.validators.validate_wireguard_public_key], verbose_name="Peer's Public Key")),
                ('dns', models.TextField(blank=True, help_text='Comma separated list.', validators=[django_wireguard.validators.validate_allowed_ips], verbose_name='DNS')),
                ('address', models.CharField(blank=True, max_length=20, validators=[django_wireguard.validators.validate_private_ipv4], verbose_name='Address')),
                ('interface_allowed_ips', models.TextField(blank=True, help_text='One per line', validators=[django_wireguard.validators.validate_allowed_ips], verbose_name='Interface Allowed IPs')),
                ('allowed_ips', models.TextField(blank=True, help_text='Comma separated list.', validators=[django_wireguard.validators.validate_allowed_ips], verbose_name='Allowed IPs')),
                ('interface_persistent_keepalive', models.PositiveIntegerField(blank=True, default=0, verbose_name='Interface Persistent Keepalive')),
                ('persistent_keepalive', models.PositiveIntegerField(blank=True, default=0, verbose_name='Persistent Keepalive')),
                ('interface', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='peers', to='django_wireguard.WireguardInterface', verbose_name='Interface')),
            ],
            options={
                'verbose_name': 'WireGuard Peer',
                'verbose_name_plural': 'WireGuard Peers',
                'unique_together': {('interface', 'name')},
            },
        ),
    ]



"""
Default settings for `django_wireguard` package.
"""
from django.conf import settings

WIREGUARD_ENDPOINT = getattr(settings, 'WIREGUARD_ENDPOINT', 'localhost')
"""The endpoint for the peer configuration. Set it to the server Public IP address or domain."""

WIREGUARD_STORE_PRIVATE_KEYS = getattr(settings, 'WIREGUARD_STORE_PRIVATE_KEYS', True)
"""Set this option to False to prevent storing peer private keys in the database."""

WIREGUARD_WAGTAIL_SHOW_IN_SETTINGS = getattr(settings, 'WIREGUARD_WAGTAIL_SHOW_IN_SETTINGS', True)
"""Set this to False to show WireGuard models in root sidebar instead of settings panel."""



from django.core.management.base import BaseCommand, CommandError
from django_wireguard.wireguard import PrivateKey, PublicKey

from django_wireguard import settings
from django_wireguard.models import WireguardPeer, WireguardInterface


class Command(BaseCommand):
    help = 'Create a WireGuard Peer'

    def add_arguments(self, parser):
        parser.add_argument('interface', type=str,
                            help="interface's name")
        parser.add_argument('name', type=str,
                            help="peer's name")
        parser.add_argument('--address', nargs='*', type=str,
                            help="specify the address for the peer.")
        parser.add_argument('--dns', nargs='*', type=str,
                            help="specify DNS for the peer.")
        parser.add_argument('--allowed-ips', nargs='*', type=str,
                            help="specify AllowedIPs for the peer.")
        parser.add_argument('--interface-allowed-ips', nargs='*', type=str,
                            help="specify Interface AllowedIPs for the peer.")
        parser.add_argument('--persistent-keepalive', nargs='?', type=int, default=0,
                            help="specify PersistentKeepalive for the peer.")
        parser.add_argument('--interface-persistent-keepalive', nargs='?', type=int, default=0,
                            help="specify Interface PersistentKeepalive for the peer.")

        group = parser.add_mutually_exclusive_group()
        group.add_argument('--private-key', nargs='?', type=str, help="specify the private key for the peer.")
        group.add_argument('--public-key', nargs='?', type=str, help="specify the public key for the peer.")

    def handle(self, *args, **options):
        name = options['name']
        try:
            interface = WireguardInterface.objects.get(name=options['interface'])
        except WireguardInterface.DoesNotExist:
            raise CommandError("Requested interface does not exist.")

        if options['private_key'] is None and options['public_key'] is None:
            private_key = PrivateKey.generate()
            public_key = private_key.public_key()
        elif options['public_key'] is None:
            try:
                private_key = PrivateKey(options['private_key'])
            except ValueError:
                raise CommandError("Invalid Private Key.")
            public_key = str(private_key.public_key())
        else:
            try:
                public_key = str(PublicKey(options['public_key']))
            except ValueError:
                raise CommandError("Invalid Public Key.")
            private_key = None

        if WireguardPeer.objects.filter(public_key=public_key).exists():
            raise CommandError("A peer with the same key already exists.")

        peer = WireguardPeer(name=name,
                             interface=interface,
                             private_key=private_key and str(private_key),
                             public_key=str(public_key),
                             address=','.join(options['address'] or []),
                             dns=','.join(options['dns'] or []),
                             allowed_ips=','.join(options['allowed_ips'] or []),
                             interface_allowed_ips=','.join(options['interface_allowed_ips'] or []),
                             persistent_keepalive=options['persistent_keepalive'],
                             interface_persistent_keepalive=options['interface_persistent_keepalive'])

        if not settings.WIREGUARD_STORE_PRIVATE_KEYS and private_key is not None:
            self.stderr.write(self.style.NOTICE("Removing Private Key references before adding peer."))
            peer.private_key = None

        # save peer before configuration to get it an IP address
        peer.save()

        # if peer didn't save private key and it's available, temporarily add it to generate conf with it
        if peer.private_key is None and private_key is not None:
            peer.private_key = private_key

        self.stderr.write(self.style.SUCCESS("\nPeer configuration:"
                                             "\n-------------------\n"))
        self.stdout.write(peer.get_config())
        self.stderr.write('\n\n')

        self.stderr.write(self.style.SUCCESS(f"Peer added successfully: {name}.\n"))



from django.contrib import admin
from django.utils.safestring import mark_safe

from django_wireguard.models import WireguardPeer, WireguardInterface
from django_wireguard.forms import WireguardPeerForm


class WireguardPeerInlineAdmin(admin.StackedInline):
    model = WireguardPeer
    show_change_link = True
    extra = 1


@admin.register(WireguardInterface)
class WireguardInterfaceAdmin(admin.ModelAdmin):
    model = WireguardInterface
    list_display = ('name', 'address', 'listen_port', 'public_key')
    inlines = [WireguardPeerInlineAdmin]


@admin.register(WireguardPeer)
class WireguardPeerAdmin(admin.ModelAdmin):
    model = WireguardPeer
    form = WireguardPeerForm
    change_form_template = 'django_wireguard/wireguardpeer_change_form.html'
    list_display = ('name', 'address', 'public_key')
    list_filter = ()

    def config(self, obj):
        return mark_safe(f'<pre>{obj.get_config()}</pre>')
    config.short_description = 'Config'



from django.apps import AppConfig
from django.db import DatabaseError


class DjangoWireguardConfig(AppConfig):
    name = 'django_wireguard'
    verbose_name = "Django WireGuard"

    def ready(self):
        """
        Sync WireGuard interfaces with system's.

        Does nothing if any DatabaseError arises.
        """
        from django_wireguard.sync_wg import sync_wireguard_interfaces
        try:
            sync_wireguard_interfaces()
        except DatabaseError:
            pass


class DjangoWireguardWagtailConfig(AppConfig):
    name = 'django_wireguard.wagtail'
    label = 'django_wireguard_wagtail'
    verbose_name = "Django WireGuard Wagtail Integration"



import base64
import ipaddress
from enum import Enum
from typing import Optional, List, Union

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from pyroute2 import WireGuard as PyRouteWireGuard, IPRoute


class PublicKey:
    """
    X25519 PublicKey Object.

    Handles validation and base64 conversion.

    :param public_key: base64 encoded key or X25519PublicKey object
    :type public_key: Union[str, X25519PublicKey]
    """
    __slots__ = ('__public_key',)

    def __init__(self, public_key: Union[str, X25519PublicKey]):
        if isinstance(public_key, X25519PublicKey):
            self.__public_key = public_key
        elif isinstance(public_key, str):
            self.__public_key = X25519PublicKey.from_public_bytes(base64.b64decode(public_key))
        else:
            raise TypeError("public_key must be a string or X25519PublicKey object")

    def __str__(self):
        value = base64.b64encode(
            self.__public_key.public_bytes(serialization.Encoding.Raw,
                                           serialization.PublicFormat.Raw))
        if isinstance(value, bytes):
            value = value.decode('ascii')
        return value


class PrivateKey:
    """
    X25519 PrivateKey Object.

    Handles validation and base64 conversion.

    :param private_key: base64 encoded key or X25519PublicKey object
    :type private_key: Union[str, X25519PublicKey]
    """
    __slots__ = ('__private_key',)

    def __init__(self, private_key: Union[str, X25519PrivateKey]):
        if isinstance(private_key, X25519PrivateKey):
            self.__private_key = private_key
        elif isinstance(private_key, str):
            self.__private_key = X25519PrivateKey.from_private_bytes(base64.b64decode(private_key))
        else:
            raise TypeError("private_key must be a string or X25519PrivateKey object")

    @classmethod
    def generate(cls):
        """
        Generate a new PrivateKey object.

        :return: Newly generated PrivateKey object
        :rtype: PrivateKey
        """
        return cls(X25519PrivateKey.generate())

    def __str__(self):
        value = base64.b64encode(
            self.__private_key.private_bytes(serialization.Encoding.Raw,
                                             serialization.PrivateFormat.Raw,
                                             serialization.NoEncryption()))
        if isinstance(value, bytes):
            value = value.decode('ascii')
        return value

    def public_key(self) -> PublicKey:
        """
        Return associated PublicKey

        :return: Associated PublicKey
        :rtype: PublicKey
        """
        return PublicKey(self.__private_key.public_key())


class WireGuardException(Exception):
    """
    Exception for WireGuard Interface operational errors (usually wraps NetLink Exceptions).
    """


class WireGuard:
    """
    WireGuard Interface abstraction class.

    This class wraps :mod:`pyroute2` methods to manage a WireGuard interface via NetLink.

    :param interface_name: WireGuard Interface Name
    """
    __slots__ = ('__ifname', '__ifindex')
    __wg = None
    __ipr = None

    class ErrorCode(Enum):
        """
        Enumeration of useful NetLink errors.

        Use this enum to check for matching NetLinkException codes.
        """
        NO_SUCH_DEVICE = 19

    def __init__(self, interface_name):
        self.__connect_backend()
        self.__ifname = interface_name
        interface = self.__get_interface_index(interface_name)
        if not interface:
            raise WireGuardException("Interface does not exist.")
        self.__ifindex = interface

    @classmethod
    def __connect_backend(cls):
        """
        Initialize static pyroute backend objects.

        These attributes are made static such that they won't clog up the file descriptors
        whenever too many are instantiated and not garbage collected.
        """
        if cls.__wg is None:
            cls.__wg = PyRouteWireGuard()
        if cls.__ipr is None:
            cls.__ipr = IPRoute()

    @classmethod
    def create_interface(cls, interface_name: str) -> 'WireGuard':
        """
        Create and enable (set state up) a new WireGuard interface.

        :param interface_name: WireGuard interface name
        :type interface_name: str
        :return: WireGuard object handling the newly created interface
        :rtype: WireGuard
        """
        cls.__connect_backend()
        cls.__ipr.link('add', ifname=interface_name, kind='wireguard')
        cls.__ipr.link('set', index=cls.__get_interface_index(interface_name), state='up')
        return cls(interface_name)

    @classmethod
    def get_or_create_interface(cls, interface_name: str) -> ('WireGuard', bool):
        """
        Get or create a new WireGuard interface.

        Create and enable (set state up) a new WireGuard interface.
        If this already exists, get it.

        :param interface_name: WireGuard interface name
        :type interface_name: str
        :return: WireGuard object handling the interface, whether the interface has been created
        :rtype: (WireGuard, bool)
        """
        cls.__connect_backend()
        interface = cls.__get_interface_index(interface_name)
        if not interface:
            return cls.create_interface(interface_name), True
        return cls(interface_name), False

    @classmethod
    def get_interface(cls, interface_name: str) -> Optional['WireGuard']:
        """
        Same as calling constructor, but returns None instead of raising WireGuardException

        :param interface_name: WireGuard interface name
        :return: WireGuard object handling the interface or None
        :rtype: WireGuard, optional
        """
        cls.__connect_backend()
        try:
            return cls(interface_name)
        except WireGuardException:
            return None

    @classmethod
    def __get_interface_index(cls, interface_name: str) -> Optional[int]:
        """
        Get iproute device index by interface name

        :param interface_name: WireGuard interface name
        :return: IPRoute device index
        :rtype: int, optional
        """
        cls.__connect_backend()
        interface: list = cls.__ipr.link_lookup(ifname=interface_name)
        if not interface:
            return None
        return interface[0]

    @property
    def interface_name(self):
        """
        The interface name.

        :return: interface name
        :rtype: str
        """
        return self.__ifname

    def get_ip_addresses(self) -> List[str]:
        """
        Return the interface's addresses.

        :return: list of IPv4 interfaces (IP with CIDR) as str
        :rtype: list
        """
        interface_data = self.__ipr.get_addr(label=self.interface_name)
        return list(map(
            lambda i: dict(i['attrs'])['IFA_ADDRESS'] + '/' + str(i['prefixlen']),
            interface_data
        ))

    def set_ip_addresses(self, *ip_addresses: str):
        """
        Set the interface's addresses.

        :param ip_addresses: new addresses
        """
        old_ip_addresses = self.get_ip_addresses()
        new_ip_addresses = []
        for address in ip_addresses:
            try:
                address = ipaddress.IPv4Interface(address)
            except Exception as e:
                raise ValueError(e)

            new_ip_addresses.append(str(address))

        for address in old_ip_addresses:
            ip, mask = address.split('/')
            if address not in new_ip_addresses:
                self.__ipr.addr('del', self.__ifindex,
                                address=ip, mask=int(mask))

        for address in new_ip_addresses:
            ip, mask = address.split('/')
            if address not in old_ip_addresses:
                self.__ipr.addr('add', self.__ifindex,
                                address=ip, mask=int(mask))

    def set_interface(self, **kwargs):
        """
        Set interface parameters.

        :param kwargs: :mod:`pyroute2.WireGuard.set` kwargs https://docs.pyroute2.org/wireguard.html
        """
        self.__wg.set(self.__ifname, **kwargs)

    def set_peer(self, public_key, *allowed_ips, **kwargs):
        """
        Set a peer on the interface

        :param public_key: peer's public key
        :param allowed_ips: peer's AllowedIPs
        :param kwargs: peer struct kwargs https://docs.pyroute2.org/wireguard.html
        """
        self.set_interface(peer={
            'public_key': str(public_key),
            'allowed_ips': [str(ipaddress.IPv4Interface(ip)) for ip in allowed_ips],
            **kwargs
        })

    def set_peers(self, *peers):
        """
        Set multiple peers

        :param peers: peer structs https://docs.pyroute2.org/wireguard.html
        """
        for peer in peers:
            self.set_interface(peer=peer)

    def remove_peers(self, *public_keys):
        """
        Remove peers by public key

        :param public_keys: peers' public keys
        """
        for pubkey in public_keys:
            peer = {'public_key': str(pubkey), 'remove': True}
            self.set_interface(peer=peer)

    def delete(self):
        """
        Delete the WireGuard interface.

        Set state down and remove device.
        """
        self.__ipr.link('set', index=self.__ifindex, state='down')
        self.__ipr.link('delete', index=self.__ifindex)
